--- old/usr/src/uts/common/fs/vfs.c	Fri Jul 19 18:39:52 2013
+++ new/usr/src/uts/common/fs/vfs.c	Fri Jul 19 18:39:52 2013
@@ -84,6 +84,8 @@
 #include <sys/spa.h>
 #include <sys/lofi.h>
 #include <sys/bootprops.h>
+#include <sys/fsh.h>
+#include <sys/fsh_impl.h>
 
 #include <vm/page.h>
 
@@ -216,13 +218,13 @@
 int
 fsop_mount(vfs_t *vfsp, vnode_t *mvp, struct mounta *uap, cred_t *cr)
 {
-	return (*(vfsp)->vfs_op->vfs_mount)(vfsp, mvp, uap, cr);
+	return (fsh_mount(vfsp, mvp, uap, cr));
 }
 
 int
 fsop_unmount(vfs_t *vfsp, int flag, cred_t *cr)
 {
-	return (*(vfsp)->vfs_op->vfs_unmount)(vfsp, flag, cr);
+	return (fsh_unmount(vfsp, flag, cr));
 }
 
 int
@@ -229,7 +231,7 @@
 fsop_root(vfs_t *vfsp, vnode_t **vpp)
 {
 	refstr_t *mntpt;
-	int ret = (*(vfsp)->vfs_op->vfs_root)(vfsp, vpp);
+	int ret = fsh_root(vfsp, vpp);
 	/*
 	 * Make sure this root has a path.  With lofs, it is possible to have
 	 * a NULL mountpoint.
@@ -247,7 +249,7 @@
 int
 fsop_statfs(vfs_t *vfsp, statvfs64_t *sp)
 {
-	return (*(vfsp)->vfs_op->vfs_statvfs)(vfsp, sp);
+	return (fsh_statfs(vfsp, sp));
 }
 
 int
@@ -274,7 +276,7 @@
 	    fidp->fid_len == XATTR_FIDSZ)
 		return (xattr_dir_vget(vfsp, vpp, fidp));
 
-	return (*(vfsp)->vfs_op->vfs_vget)(vfsp, vpp, fidp);
+	return (fsh_vget(vfsp, vpp, fidp));
 }
 
 int
@@ -528,6 +530,7 @@
 	vfsimpl_setup(vfsp);
 	vfsp->vfs_data = (data);
 	vfs_setops((vfsp), (op));
+	vfsp->vfs_fshrecord = fsh_fsrec_create();
 }
 
 /*
@@ -1785,6 +1788,7 @@
 		}
 		/* Return vfsp to caller. */
 		*vfspp = vfsp;
+		fsh_exec_create_callbacks(vfsp);
 	}
 errout:
 	vfs_freeopttbl(&mnt_mntopts);
@@ -4319,6 +4323,10 @@
 		vfsp->vfs_femhead = NULL;
 	}
 
+	/* FSH cleanup */
+	fsh_fsrec_destroy(vfsp->vfs_fshrecord);
+	vfsp->vfs_fshrecord = NULL;
+
 	if (vfsp->vfs_implp)
 		vfsimpl_teardown(vfsp);
 	sema_destroy(&vfsp->vfs_reflock);
@@ -4345,6 +4353,7 @@
 {
 	ASSERT(vfsp->vfs_count != 0);
 	if (atomic_add_32_nv(&vfsp->vfs_count, -1) == 0) {
+		fsh_exec_destroy_callbacks(vfsp);
 		VFS_FREEVFS(vfsp);
 		lofi_remove(vfsp);
 		if (vfsp->vfs_zone)
